/*
 * Cluster Information Service - a monitoring system for Linux clusters
 * Copyright (C) 2000 Institute of Informatics, Slovak Academy of Sciences.
 * Written by Jan Astalos (astalos.ui@savba.sk)
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Kernel module for monitoring of network devices with start/stop notification
 * via netlink device.
 */

#include <linux/config.h>
#include <linux/module.h>

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <asm/uaccess.h>

#include <net/inet_common.h>
#include <net/tcp.h>
#include <linux/netdevice.h>
#include <linux/proc_fs.h>

#include "cis.h"
#include "cis_mon.h"

static struct sock *nmsk;

int netdevlist_read_proc(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
        int len = 0;
        off_t pos = 0;
	off_t begin = 0;
        struct cis_netdevinfo msg;
        struct net_device *dev;
        struct net_device_stats *stats;

	read_lock(&dev_base_lock);
	for (dev = dev_base; dev != NULL; dev = dev->next) {

                if (dev == &loopback_dev)
                        continue;

                stats = (dev->get_stats ? dev->get_stats(dev): NULL);

                if (!stats)
                        continue;

                pos += sizeof(struct cis_netdevinfo);
                if (pos < offset)
                        continue;

		msg.status = (dev->flags&IFF_UP) ? 1 : 0;
		strncpy(msg.name, dev->name, IFNAMSIZ);
                msg.receive = stats->rx_bytes >> 10;
                msg.transmit = stats->tx_bytes >> 10;
                msg.collisions = stats->collisions;
                

                memcpy(page + len, &msg, sizeof(struct cis_netdevinfo));
                len += sizeof(struct cis_netdevinfo);
                if(len >= count)
			break;
	}
	read_unlock(&dev_base_lock);
	
        begin = len - (pos - offset);
        *start = page + begin;
	len -= begin;
	if (len > count) len = count;
	if (len <= count) *eof = 1;
	if (len < 0) len = 0;

	return len;
}


static int netmon_msg(struct net_device *dev, unsigned char action)
{
        struct cis_netdevinfo *msg;
        size_t len =  sizeof(struct cis_netdevinfo);
        struct net_device_stats *stats = (dev->get_stats ? dev->get_stats(dev): NULL);

        struct sk_buff *outskb = alloc_skb(len, GFP_ATOMIC);

        if (outskb) {
                skb_put(outskb, len);
                msg = (struct cis_netdevinfo  *) outskb->data;

		msg->status = (dev->flags&IFF_UP) ? 1 : 0;
                strncpy (msg->name, dev->name, IFNAMSIZ);

                msg->receive    = (stats) ? stats->rx_bytes >> 10 : 0;
                msg->transmit   = (stats) ? stats->tx_bytes >> 10 : 0;
		msg->collisions = (stats) ? stats->collisions : 0;

                netlink_broadcast(nmsk, outskb, 0, ~0, GFP_KERNEL);
        }
        else
                return (-1);

	return (0);
}

static int netmon_notifier(struct notifier_block *this, unsigned long msg, void *data)
{
        struct net_device *dev = (struct net_device*)data;

        switch (msg) {
        case NETDEV_DOWN:
        case NETDEV_UNREGISTER:
		netmon_msg(dev, OBJ_DESTROY);
                break;
        case NETDEV_UP:
                netmon_msg(dev, OBJ_CREATE);
                break;
        }
        return NOTIFY_DONE;
}

struct notifier_block netmon_netdev_notifier={
        netmon_notifier,
	NULL,
	0
};

static struct proc_dir_entry *netdevlist_proc_entry;

int netmon_init(void)
{
        nmsk = netlink_kernel_create(NETLINK_NETMON, NULL);
        if (nmsk == NULL) {
                printk("netmon_init: cannot initialize netlink\n");
                return -ENODEV;
        }
        
	netdevlist_proc_entry = create_proc_read_entry("cis_netdevlist", 0, NULL, netdevlist_read_proc, NULL);
	if (!netdevlist_proc_entry) {
		printk("procmon_init: cannot register netlist\n");
		return -EPERM;
	}

        register_netdevice_notifier(&netmon_netdev_notifier);

        return 0;
}

int init_module(void)
{
        return (netmon_init ());
}

void cleanup_module(void)
{
	remove_proc_entry("cis_netdevlist", 0);
	unregister_netdevice_notifier(&netmon_netdev_notifier);
	sock_release (nmsk->socket);
}
