/*
 * Cluster Information Service - a monitoring system for Linux clusters
 * Copyright (C) 2000 Institute of Informatics, Slovak Academy of Sciences.
 * Written by Jan Astalos (astalos.ui@savba.sk)
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Kernel module for monitoring of INET sockets with open/close notification
 * via netlink device.
 */

#include <linux/config.h>
#include <linux/module.h>

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <asm/uaccess.h>

#include <net/inet_common.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <linux/proc_fs.h>

#include "cis.h"
#include "cis_mon.h"

extern struct net_proto_family inet_family_ops;

struct proto_ops old_inet_stream_ops;
struct proto_ops old_inet_dgram_ops;
struct net_proto_family old_inet_family_ops;

static struct sock *smsk;

#define TCP_HDRLEN 24 + 24
#define UDP_HDRLEN 24 + 8

void copy_sock_data(struct cis_sockinfo *msg, struct sock *sk)
{
	msg->saddr.s_addr = sk->rcv_saddr;
	msg->sport        = ntohs(sk->sport);
	msg->daddr.s_addr = sk->daddr;
	msg->dport        = ntohs(sk->dport);
	msg->pid          = sk->owner_pid;
	msg->uid          = sk->owner_uid;
	msg->sent         = sk->bytes_sent;
	msg->rcvd         = sk->bytes_received;
}
int socklist_read_proc(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	int i, len = 0;
        off_t pos = 0;
        off_t begin = 0;
        struct cis_sockinfo msg;
	struct sock *sk;

	read_lock(&udp_hash_lock);
	for (i = 0; i < UDP_HTABLE_SIZE; i++) {

		for (sk = udp_hash[i]; sk; sk = sk->next) {
			if (sk->family != PF_INET)
				continue;

			pos += sizeof(struct cis_sockinfo);
			if (pos < offset)
				continue;

			msg.type = SOCK_DGRAM;
			copy_sock_data(&msg, sk);

			memcpy (page + len, &msg, sizeof(struct cis_sockinfo));
			len += sizeof(struct cis_sockinfo);
			if(len >= count) {
				read_unlock(&udp_hash_lock);
				goto out;
			}
		}
	}
	read_unlock(&udp_hash_lock);

	local_bh_disable();

	/* Established TCP sockets */
	for (i = 0; i < tcp_ehash_size; i++) {
		struct tcp_ehash_bucket *head = &tcp_ehash[i];

		read_lock(&head->lock);
		for(sk = head->chain; sk; sk = sk->next) {
			if (!TCP_INET_FAMILY(sk->family))
				continue;
			pos += sizeof(struct cis_sockinfo);
			if (pos < offset)
				continue;

			msg.type = SOCK_STREAM;
			copy_sock_data(&msg, sk);

			memcpy (page + len, &msg, sizeof(struct cis_sockinfo));
			len += sizeof(struct cis_sockinfo);
			if(len >= count) {
				read_unlock(&head->lock);
				goto out_bh;
			}
		}
		read_unlock(&head->lock);
	}
out_bh:
	local_bh_enable();
out:
	begin = len - (pos - offset);
	*start = page + begin;
	len -= begin;
	if (len > count) len = count;
	if (len <= count) *eof = 1;
	if (len < 0) len = 0;
	
	return len;
}

static int sockmon_msg(struct socket *sock, unsigned char action)
{
        struct sock *sk = sock->sk;
        struct cis_sockinfo *msg;

	struct sk_buff *outskb = alloc_skb(sizeof(struct cis_sockinfo), GFP_ATOMIC);

        if (outskb) {
                skb_put(outskb, sizeof(struct cis_sockinfo));
                msg = (struct cis_sockinfo *) outskb->data;

		msg->priv         = (void *) (int) action;
                msg->type         = sock->type;
                msg->saddr.s_addr = sk->rcv_saddr;
                msg->sport        = ntohs(sk->sport);
                msg->daddr.s_addr = sk->daddr;
                msg->dport        = ntohs(sk->dport);
                msg->pid          = sk->owner_pid;
                msg->uid          = sk->owner_uid;
                msg->sent         = sk->bytes_sent;
                msg->rcvd         = sk->bytes_received;

                netlink_broadcast(smsk, outskb, 0, ~0, GFP_KERNEL);
        }
        else
                return (-1);

	return (0);
}

/*
 * New inet create.
 */
int new_inet_create(struct socket *sock, int protocol)
{
        int err;
        struct sock *sk;
        
        err = old_inet_family_ops.create (sock, protocol);

        if (err == 0) {
                sk = sock->sk;

                sk->owner_pid      = current->pid;
                sk->owner_uid      = current->uid;
                sk->bytes_sent     = 0;
                sk->bytes_received = 0;
        }
        
        return err;
}

/*
 * Stream operations wrappers
 */

int new_stream_release(struct socket *sock)
{
	sockmon_msg (sock, OBJ_DESTROY);
	return (old_inet_stream_ops.release) (sock);
}

int new_stream_bind(struct socket *sock, struct sockaddr *uaddr, int addr_len)
{
        struct sock *sk = sock->sk;
        int err;

        err = old_inet_stream_ops.bind (sock, uaddr, addr_len);

        if (sk) {
                sk->owner_pid      = current->pid;
                sk->owner_uid      = current->uid;
                sk->bytes_sent     = 0;
                sk->bytes_received = 0;
        }
        return err;
}

int new_stream_connect(struct socket *sock, struct sockaddr * uaddr,
			int addr_len, int flags)
{
        int err = old_inet_stream_ops.connect (sock, uaddr, addr_len, flags);

        if (err == 0)
                sockmon_msg (sock, OBJ_CREATE);

        return err;
}

int new_stream_accept(struct socket *sock, struct socket *newsock, int flags)
{
        int err = old_inet_stream_ops.accept (sock, newsock, flags);
        struct sock *sk;
        
        if (err == 0) {
                sk = newsock->sk;

                sk->owner_pid      = current->pid;
                sk->owner_uid      = current->uid;
                sk->bytes_sent     = 0;
                sk->bytes_received = 0;
                
                sockmon_msg (newsock, OBJ_CREATE);
        }

        return err;
}

int new_stream_sendmsg(struct socket *sock, struct msghdr *msg, int size,
                       struct scm_cookie *scm)
{
        struct sock *sk = sock->sk;
        int err;

        err = old_inet_stream_ops.sendmsg (sock, msg, size, scm);

        if (sk && err > 0) {
                if (sk->owner_pid != current->pid) {
                        sk->owner_pid = current->pid;
                        sk->owner_uid = current->uid;
                }
		sk->bytes_sent += err + TCP_HDRLEN;
//		printk (KERN_INFO "TCP sent %d\n", err);
        }
        return err;
}

int new_stream_recvmsg(struct socket *sock, struct msghdr *msg, int size,
                       int flags, struct scm_cookie *scm)
{
        struct sock *sk = sock->sk;
        int err;

        err = old_inet_stream_ops.recvmsg (sock, msg, size, flags, scm);

        if (sk && err > 0) {
                if (sk->owner_pid != current->pid) {
                        sk->owner_pid = current->pid;
                        sk->owner_uid = current->uid;
                }
                sk->bytes_received += err + TCP_HDRLEN;
//		printk (KERN_INFO "TCP rcvd %d\n", err);
        }
        return err;
}
/*
 * Datagram operations wrappers.
 * Yes. Some functions are duplicated. But what if the original handler will
 * not be the same in the future...
 */

int new_dgram_release(struct socket *sock)
{
        sockmon_msg (sock, OBJ_DESTROY);
        return (old_inet_dgram_ops.release) (sock);
}

int new_dgram_bind(struct socket *sock, struct sockaddr *uaddr, int addr_len)
{
        struct sock *sk = sock->sk;
        int err;

        err = old_inet_dgram_ops.bind (sock, uaddr, addr_len);

        if (sk) {
                sk->owner_pid      = current->pid;
                sk->owner_uid      = current->uid;
                sk->bytes_sent     = 0;
                sk->bytes_received = 0;
        }
        return err;
}

int new_dgram_connect(struct socket *sock, struct sockaddr * uaddr,
			int addr_len, int flags)
{
        struct sock *sk = sock->sk;
        int err = old_inet_dgram_ops.connect (sock, uaddr, addr_len, flags);

        if (sk && err == 0) {
                sk->owner_pid      = current->pid;
                sk->owner_uid      = current->uid;
                sk->bytes_sent     = 0;
                sk->bytes_received = 0;
                
                sockmon_msg (sock, OBJ_CREATE);
        }

        return err;
}

int new_dgram_sendmsg(struct socket *sock, struct msghdr *msg, int size,
                       struct scm_cookie *scm)
{
        struct sock *sk = sock->sk;
        int err;

        err = old_inet_dgram_ops.sendmsg (sock, msg, size, scm);
        
        if (sk && err > 0) {
                if (sk->owner_pid != current->pid) {
                        sk->owner_pid = current->pid;
                        sk->owner_uid = current->uid;
                }
                sk->bytes_sent += err + UDP_HDRLEN;
//		printk (KERN_INFO "UDP sent %d\n", err);
        }
        return err;
}

int new_dgram_recvmsg(struct socket *sock, struct msghdr *msg, int size,
                       int flags, struct scm_cookie *scm)
{
        struct sock *sk = sock->sk;
        int err;

        err = old_inet_dgram_ops.recvmsg (sock, msg, size, flags, scm);

        if (sk && err > 0) {
                if (sk->owner_pid != current->pid) {
                        sk->owner_pid = current->pid;
                        sk->owner_uid = current->uid;
                }
		sk->bytes_received += err + UDP_HDRLEN;
//		printk (KERN_INFO "UDP rcvd %d\n", err);
        }
        return err;
}

static struct proc_dir_entry *socklist_proc_entry;

int sockmon_init(void)
{
        smsk = netlink_kernel_create(NETLINK_SOCKMON, NULL);
        if (smsk == NULL) {
                printk("sockmon_init: cannot initialize netlink\n");
                return -ENODEV;
        }
        
	socklist_proc_entry = create_proc_read_entry("cis_socklist", 0, NULL, socklist_read_proc, NULL);
	if (!socklist_proc_entry) {
		printk("procmon_init: cannot register socklist\n");
		return -EPERM;
	}
        
        old_inet_stream_ops = inet_stream_ops;
        old_inet_dgram_ops  = inet_dgram_ops;
        old_inet_family_ops = inet_family_ops;
	
        inet_stream_ops.release = new_stream_release;
        inet_stream_ops.bind    = new_stream_bind;
        inet_stream_ops.connect = new_stream_connect;
        inet_stream_ops.accept  = new_stream_accept;
        inet_stream_ops.sendmsg = new_stream_sendmsg;
        inet_stream_ops.recvmsg = new_stream_recvmsg;

        inet_dgram_ops.release = new_dgram_release;
        inet_dgram_ops.bind    = new_dgram_bind;
        inet_dgram_ops.connect = new_dgram_connect;
        inet_dgram_ops.sendmsg = new_dgram_sendmsg;
        inet_dgram_ops.recvmsg = new_dgram_recvmsg;

        inet_family_ops.create = new_inet_create;

//        printk (KERN_INFO "Socket monitoring enabled.\n");

        return 0;
}

int init_module(void)
{
        return (sockmon_init ());
}

void cleanup_module(void)
{
        inet_stream_ops = old_inet_stream_ops;
        inet_dgram_ops  = old_inet_dgram_ops;
        inet_family_ops = old_inet_family_ops;

        remove_proc_entry("cis_socklist", 0);
        sock_release (smsk->socket);

//	printk (KERN_INFO "Socket monitoring disabled.\n");
}
