/*
 * Cluster Information Service - a monitoring system for Linux clusters
 * Copyright (C) 2000 Institute of Informatics, Slovak Academy of Sciences.
 * Written by Jan Astalos (astalos.ui@savba.sk)
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * CIS server daemon - core code with interface to monitors.
 */

#include <unistd.h>
#include <signal.h>
#include <locale.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <netinet/in.h>

#include "cis.h"
#include "cis_xdr.h"
#include "cisd.h"

#define DEF_OPEN_MODE             O_CREAT|O_APPEND|O_NOCTTY

#ifndef DEF_INTERVAL
#define DEF_INTERVAL              1.
#endif

#ifndef DEF_MSG_COUNT
#define DEF_MSG_COUNT             60
#endif

#define SOCKFLAGS     (MSG_DONTWAIT | MSG_NOSIGNAL)

#define allocCPU(a,b,c) (((a) * ((10000 * (b)) / (c))) / 10000)

#define TMP_FLAG     (1<<5)

#define is_marked(a) (a->flags & TMP_FLAG)
#define mark(a)      {a->flags |= TMP_FLAG;}
#define unmark(a)    {a->flags &= ~TMP_FLAG;}


extern void dispatcher(struct svc_req *rqstp, SVCXPRT *transp);
extern void daemonize (void);
extern uint create_auth_code (void);

struct sockaddr_in srv_addr;
struct in_addr myaddr;

struct cis_info my_info;

int prognum = CIS_PROG;

int nhosts;
struct host_info **hosttab, *hosts_by_addr[256];

struct timeval timeout={5,0};
struct timeval current_tv;

struct monitor_msg monbuff;
char *buffptr = monbuff.data;
#define bufflen (buffptr - (char *) &monbuff)
unsigned long msg_time;

short msglen;
int monitor_socket;
float interval = DEF_INTERVAL;
struct monitor_reply reply;

char *config_file = NULL;

char **perf_types = NULL;
int nperf_types = 0;

extern int (*dispatch_fn[NMONITORS][NMSGTYPES])(struct host_info *, int);
void parse_config(void);

void parse_args(int argc, char **argv)
{
        int c;

	while ((c = getopt(argc, argv, "i:c:")) != EOF) {

                switch (c) {
                case 'i':
                        interval = atof(optarg);
                        if (interval == 0.) {
                                syslog(LOG_WARNING, "interval cannot be set to zero. reset to default %5.2f [s]", DEF_INTERVAL);
                                interval = DEF_INTERVAL;
                        }
                        break;
                case 'c':
                        config_file = optarg;
			break;
                }
        }
}

void free_host(struct host_info *host)
{
	hash_destroy(host->proc_ht);
	hash_destroy(host->sock_ht);
        free(host);
}

int monitor_idle(struct monitor_info *monitor)
{
//	return (monitor_time(monitor) > (monitor->local_time + 2*monitor->clk_interval));

	return (tvaldiff (current_tv, monitor->time_tv) > 3*interval);
}

unsigned long count_rate(struct history *h, unsigned long time, unsigned long interval, unsigned long cticks)
{
        unsigned long sum = 0;
        unsigned long bound = time - interval;
        struct history_value *hp;

	if (!h->last)
		return 0;
	
        if (h->last->time < bound)
                return 0;

        if (h->last->time - h->interval > bound)
                return (h->sum / (((float) (time - h->last->time + h->interval))
                        / cticks));

        for (hp = h->last; hp->time >= bound;) {

                sum += hp->change;

                if (--hp < h->history)
                        hp = h->history + HISTORY_LENGTH - 1;

                if (hp == h->last)
                        break;
        }

        return (sum / ((float) interval / cticks));
}


/*
 * Process related functions
 */
/*int slice (struct host_info *host, int niceval)
{
        int slice = niceval;
        
        if (niceval < 0)
                slice = -niceval;
        if (slice > 20)
                slice = 20;
        slice = (slice * host->defprio + 10) / 20 + host->defprio;

        if (niceval >= 0) {
                slice = 2*host->defprio - slice;
                if (!slice)
                        slice = 1;
        }
        return (slice);
        } */

int count_pCPU(struct host_info *host, struct process_info *p, unsigned long current_time)
{
	int pCPU = 0;
	struct history *hist = &p->CPU_history;
	unsigned long interval, sum;

        /*
         * No usage.
         */
        if (!hist->interval)
		goto out;

        history_sum(hist, PROC_HISTORY_LENGTH, &interval, &sum);

        /*
         * No change during 2*'interval', process is sleeping.
         */
	if (current_time > hist->last->time + 2*interval)
                goto out;

        /*
         * Hold the last usage during 'interval' after last change.
         */
        pCPU  = (10000 * sum) / interval;

        if (current_time < hist->last->time + interval)
                goto out;

        /*
         * Between, usage will fall down to zero.
         */
        pCPU *= hist->last->time + 2*interval - current_time;
        pCPU /= interval;

    out:
        return pCPU;
}

int availableCPU(struct host_info *host)
{
	struct process_info *p;
        struct cis_procinfo *pi;
        struct history *hist;
        struct monitor_info *monitor = host->monitor[PROCMON];
        unsigned long psum_tmp, slsum_tmp, available_tmp;
        unsigned long clk;
        unsigned long available = 0;
        unsigned long min_interval;
        unsigned long psum = 0;
        unsigned long slsum = 0;
        int n = 0;

        if (!monitor)
                return 0;

        if (monitor_idle(monitor)) /* monitor died without message */
		return 0;

        clk = monitor_time(monitor);
	min_interval = 2*monitor->clk_interval;

	for_each_process (host, p) {
		pi = &p->data;
		hist = &p->CPU_history;
                n++;
		unmark(p);
                
                /* What CPU usage process has ? */
                if (hist->interval) {
                        pi->pCPU = count_pCPU(host, p, clk);
                        slsum += p->slice;
                        psum += pi->pCPU;
			mark(p);
                        
                        /*
                         * Minimal process interval is needed for deciding whether
                         * the CPU was busy all the time or it was idle some time.
                         */
                        if (hist->interval < min_interval)
                                min_interval = hist->interval;
                        continue;
                }
                pi->pCPU = 0;

		{
			unsigned long diff;
			
			diff = clk - p->CPU_history.last->time;
			if (diff < 2*p->init_interval) {
				pi->pCPU = 10000;

				if (diff > p->init_interval) {
					pi->pCPU *= diff / p->init_interval;
					pi->pCPU += p->init_interval;
				}
			}
		}
        }

        host->hostinfo.procnum = n;

        if (!n)
                return (0);

        if (!psum)
                return (10000);

        if (host->idle_at < monitor->local_time - min_interval && psum < 10000) {
                /*
                 * If CPU was not idle and there was terminated process
                 * (i.e. pCPU sum is < 100 %), we must recompute
                 * needs of all 'greedy' processes.
                 */
                available = 10000 - psum;

                slsum_tmp = 0; psum_tmp = psum;
		for_each_process (host, p) {
			if (!is_marked(p)) continue;

                        if (p->data.pCPU < allocCPU(psum, p->slice, slsum)) {
				psum_tmp -= p->data.pCPU;
                                unmark(p);
                        }
                        else
                                slsum_tmp += p->slice;

                }
                slsum = slsum_tmp;
                psum = psum_tmp;
                
		for_each_process (host, p) {
			if (!is_marked(p)) continue;

			p->data.pCPU = allocCPU(psum + available, p->data.pCPU, psum);
                        unmark(p);
		}
	}

        /*
         * Estimate how much CPU time a process with default priority
         * can get (default response time).
         */
        slsum = 20 + host->slsum;
        available = 20000;
        available_tmp = 10000;

        while (available > available_tmp) {

                available = available_tmp;
                slsum_tmp = 20;

		for_each_process (host, p) {
			if (!p->data.pCPU) continue;

			if (p->data.pCPU < allocCPU(available, p->slice, slsum)) {
				available_tmp -= p->data.pCPU;
                                p->data.pCPU = 0;
                        } else
				slsum_tmp += p->slice;
                }
                slsum = slsum_tmp;
        }

        available = allocCPU(available, 20, slsum);

        return (available);
}


struct process_info *proc_add(struct host_info *host, struct cis_procinfo *p)
{
	struct process_info *process, *q;
	struct monitor_info *monitor = host->monitor[PROCMON];
	int interval = 0;
        unsigned long time;
        
	process = calloc(1, sizeof (struct process_info));
	if (!process)
		return NULL;

	if (p) memcpy(&process->data, p, sizeof (struct cis_procinfo));

        time = process->data.start_time;
        if (!time)
                time = monitor->local_time;
        //time = monitor->local_time - monitor->clk_interval;

        /* Recompute start_time to GMT time. It's ugly but we have to cope
         * with possibly very high unsigned longs */
        process->GMT_start_time = monitor->start_tv.tv_sec;
        if (time < monitor->start_clk)
                process->GMT_start_time -= (monitor->start_clk - time)/monitor->ct_per_sec;
        else
		process->GMT_start_time += (time - monitor->start_clk)/monitor->ct_per_sec;
         
        /* Process history initialization. */
        init_history(&process->CPU_history,    time);
        init_history(&process->majflt_history, time);
        init_history(&process->read_history,   time);
        init_history(&process->write_history,  time);

        process->flags = CONSISTENT;

	list_insert (host->proclist, process);

        /* How much to wait for first usage ? */
        if (host->slsum)
		for_each_process (host, q) {
			q->data.pCPU = count_pCPU(host, q, monitor->local_time);
                        /* Compute time to wait for CPU usage (i.e. the sum of
                         * slices of all CPU intensive processes with higher
                         * priority than the new process has */
                        if (q->slice >= process->slice &&
                            q->data.pCPU > ((10000 * q->slice) / host->slsum))
				interval += q->slice;
                }

        if (interval < monitor->clk_interval)
                process->init_interval = monitor->clk_interval;
        else
                process->init_interval = interval;

	process->slice = 20 - p->nice;
	
	host->slsum += process->slice;

	hash_insert(host->proc_ht, process);
        
	return process;
}

/*
 * When parent exits, ppid for all children should be set to 1 (init).
 * In order to keep the process tree consistent.
 */
void proc_release_children(struct host_info *host, int pid)
{
	struct process_info *p;

	for_each_process (host, p)
		if (p->data.ppid == pid)
			p->data.ppid = 1;
}

void proc_remove(struct host_info *host, struct process_info *process)
{
        if (!process) return;
        
	list_remove (host->proclist, process);

	host->slsum -= process->slice;
        proc_release_children(host, process->data.pid);
	hash_remove(host->proc_ht, &process->data.pid);

	free(process);
}

void proc_update(struct host_info *host, struct process_info *p, struct cis_procinfo *newp)
{
	struct cis_procinfo *oldp = &p->data;
	struct monitor_info *monitor = host->monitor[PROCMON];
        unsigned short ctps = host->hostinfo.ct_per_sec;
	unsigned long change;

        if (oldp->nice != newp->nice) {
                host->slsum += 20 - newp->nice - p->slice;
                p->slice = 20 - newp->nice;
        }

        change = newp->utime - oldp->utime + newp->stime - oldp->stime;
	if (change) {
		history_add(&p->CPU_history, monitor->local_time, change);
		
		p->utime_sec = newp->utime/ctps;
		p->stime_sec = newp->stime/ctps;
	}

	history_add(&p->majflt_history, monitor->local_time, newp->majflt   - oldp->majflt);
	history_add(&p->read_history,   monitor->local_time, newp->disk_read - oldp->disk_read);
	history_add(&p->write_history,  monitor->local_time, newp->disk_write - oldp->disk_write);

	memcpy(oldp, newp, PROCINFO_LEN);

	p->flags |= CONSISTENT;
}

struct process_info *proc_find(struct host_info *host, unsigned int pid)
{
	if (!host->monitor[PROCMON])
		return NULL;

	return hash_find(host->proc_ht, &pid);
}


void proc_test(struct host_info *host, struct cis_procinfo *p)
{
	struct process_info *process = proc_find(host, p->pid);

        if (!process)
		proc_add(host, p);
        else
                proc_update(host, process, p);
}


/*
 * Socket related functions
 */
struct socket_info *sock_add(struct host_info *host, struct cis_sockinfo *s)
{
	struct socket_info *socket;
	unsigned long time = host->monitor[SOCKMON]->local_time;

	socket = calloc(1, sizeof (struct socket_info));
        if (!socket)
		return NULL;
	memcpy(&socket->data, s, sizeof (struct cis_sockinfo));

	list_insert (host->socklist, socket);

	init_history(&socket->sent_history, time);
        init_history(&socket->rcvd_history, time);

        socket->flags = CONSISTENT;

	hash_insert(host->sock_ht, socket);

        return socket;
}

void sock_update(struct host_info *host, struct socket_info *s, struct cis_sockinfo *news)
{
	struct cis_sockinfo *olds = &s->data;
	unsigned long time = host->monitor[SOCKMON]->local_time;

	history_add(&s->sent_history, time, news->sent - olds->sent);
	history_add(&s->rcvd_history, time, news->rcvd - olds->rcvd);

        memcpy(olds, news, SOCKINFO_LEN);

	s->flags |= CONSISTENT;
}

void sock_remove (struct host_info *host, struct socket_info *s)
{
        if(!s) return;
        
	s->flags &= ~CONSISTENT;
        hash_remove(host->sock_ht, &s->data);
}

static struct socket_info *sock_find(struct host_info *host,  struct cis_sockinfo *s)
{
	if (!host->monitor[SOCKMON])
		return NULL;

	return hash_find(host->sock_ht, s);
}

/*
 * Network device related functions
 */
static struct netdev_info *netdev_find(struct host_info *host, unsigned char id)
{
	struct netdev_info *d;

        for (d = host->netdevlist; d; d = d->next)
		if (d->data.id == id)
			return d;
	return NULL;
}

struct netdev_info *netdev_add (struct host_info *host, struct cis_netdevinfo *dev)
{
	struct netdev_info *device = calloc(1, sizeof (struct netdev_info));
	unsigned long time = host->monitor[NETMON]->local_time;

	if (!device)
		return NULL;
	memcpy(&device->data, dev, NETDEVINFO_LEN);

	init_history(&device->rxbytes_history, time);
        init_history(&device->txbytes_history, time);
	init_history(&device->collisions_history, time);

	device->flags = CONSISTENT;
	list_insert(host->netdevlist, device);

        return device;
}

void netdev_update(struct host_info *host, struct netdev_info *d, struct cis_netdevinfo *newd)
{
	struct cis_netdevinfo *oldd = &d->data;
	unsigned long time = host->monitor[NETMON]->local_time;
        
	history_add(&d->rxbytes_history, time, newd->receive - oldd->receive);
	history_add(&d->txbytes_history, time, newd->transmit - oldd->transmit);
	history_add(&d->collisions_history, time, newd->collisions - oldd->collisions);

	memcpy (oldd, newd, NETDEVINFO_LEN);
        
	d->flags |= CONSISTENT;
}

#define sys_code(el)   OFFSET(cis_hostinfo, el)

int sys_change(struct cis_hostinfo *hi, int end)
{
	unsigned char code;
        int change;

        get_object(code);

	while (code != 0xff) {
                switch (code) {
                case sys_code(procnum):       get_change(hi->procnum);       break;
                case sys_code(loads[0]):      get_change(hi->loads[0]);      break;
                case sys_code(loads[1]):      get_change(hi->loads[1]);      break;
                case sys_code(loads[2]):      get_change(hi->loads[2]);      break;
                case sys_code(totalswap):     get_change(hi->totalswap);     break;
                case sys_code(freeswap):      get_change(hi->freeswap);      break;
                case sys_code(freeram):       get_change(hi->freeram);       break;
                case sys_code(sharedram):     get_change(hi->sharedram);     break;
                case sys_code(bufferram):     get_change(hi->bufferram);     break;
                case sys_code(cachedram):     get_change(hi->cachedram);     break;
                case sys_code(idle):          get_change(hi->idle);          break;
                case sys_code(ctx_swtch):     get_change(hi->ctx_swtch);     break;
                case sys_code(swpin):         get_change(hi->swpin);         break;
                case sys_code(swpout):        get_change(hi->swpout);        break;
		case sys_code(disks_read):    get_change(hi->disks_read);    break;
		case sys_code(disks_write):   get_change(hi->disks_write);   break;
		}
		get_object(code);
        }
        return 0;
msg_error:
        return -1;
}


#define proc_code(el)   OFFSET(cis_procinfo, el)

int proc_change(struct host_info *host, unsigned int pid, int end)
{
        struct cis_procinfo proc;
        struct process_info *process = proc_find(host, pid);
        char code;
	int change;

        get_object(code);

	if (code == 1) {/* New process */
		get(&proc, PROCINFO_LEN);
		proc_test(host, &proc);
                return 0;
        }
        if (code == 2) {/* Terminated process */
                get(&proc, PROCINFO_LEN);
		proc_remove(host, process);
                return 0;
        }
        
        if (process)
		proc = process->data;
	else {
		/* Change of nonexisting process */
                memset(&proc, 0, sizeof(proc));
                proc.pid = pid;
                sprintf(proc.cmd, "unknown");
                process = proc_add(host, &proc);
        }

        while (code) {
                switch (code) {
                case proc_code(utime):      get_change(proc.utime);      break;
                case proc_code(stime):      get_change(proc.stime);      break;
                case proc_code(minflt):     get_change(proc.minflt);     break;
                case proc_code(majflt):     get_change(proc.majflt);     break;
                case proc_code(disk_read):  get_change(proc.disk_read);  break;
                case proc_code(disk_write): get_change(proc.disk_write); break;
                case proc_code(nice):       get_change(proc.nice);       break;
                case proc_code(rss):        get_change(proc.rss);        break;
                case proc_code(vm):         get_change(proc.vm);         break;
		case proc_code(ppid):       get_object(proc.ppid);       break;
		case proc_code(uid):        get_object(proc.uid);        break;
		case proc_code(cmd):        get_object(proc.cmd);        break;
		}
                get_object(code);
        }
        proc_update(host, process, &proc);

        return 0;
        
msg_error:
        return -1;
}

void sock_test(struct host_info *host, struct cis_sockinfo *s)
{
	struct socket_info *socket = sock_find(host, s);

	if (!socket)
                sock_add(host, s);
        else
                sock_update(host, socket, s);
}

#define sock_code(el)   OFFSET(cis_sockinfo, el)

int sock_change(struct host_info *host, struct cis_sockinfo *s, int end)
{
	struct cis_sockinfo sock;
	struct socket_info *socket = sock_find(host, s);
	char code;
        int change;

        get_object(code);

        if (socket)
		sock = socket->data;
        else {
                /* Change of nonexisting socket */
		memset(&sock, 0, sizeof(sock));
		memcpy(&sock, s, SOCKINFO_HDRLEN);
		socket = sock_add(host, &sock);
        }

        if (code == 1) {/* New socket */
                get(&sock + SOCKINFO_HDRLEN, SOCKINFO_LEN-SOCKINFO_HDRLEN);
                sock_update(host, socket, &sock);
                return 0;
        }
        if (code == 2) {/* Closed socket */
                get(&sock + SOCKINFO_HDRLEN, SOCKINFO_LEN-SOCKINFO_HDRLEN);
		sock_update(host, socket, &sock);
                sock_remove(host, socket);
                return 0;
        }

	while (code) {
                switch (code) {
                case sock_code(pid):      get_object(sock.pid);      break;
                case sock_code(uid):      get_object(sock.uid);      break;
                case sock_code(sent):     get_change(sock.sent);     break;
                case sock_code(rcvd):     get_change(sock.rcvd);     break;
                }
                get_object(code);
        }
        sock_update(host, socket, &sock);

        return 0;
        
msg_error:
        return -1;
}


void netdev_test(struct host_info *host, struct cis_netdevinfo *d)
{
	struct netdev_info *device = netdev_find(host, d->id);

	if (!device)
                netdev_add(host, d);
        else
                netdev_update(host, device, d);
}

#define netdev_code(el)   OFFSET(cis_netdevinfo, el)

int netdev_change(struct host_info *host, unsigned char id, int end)
{
	struct cis_netdevinfo dev;
        struct netdev_info *device = netdev_find(host, id);
	char code;
	int change;

        get_object(code);

	if (code == 1) {/* New device */
                get(&dev, NETDEVINFO_LEN);
                netdev_test(host, &dev);
                return 0;
        }
        
        if (device)
		dev = device->data;
        else {
                /* Change of nonexisting device */
                memset(&dev, 0, sizeof(dev));
		dev.id = id;
		sprintf(dev.name, "unknown");
		device = netdev_add(host, &dev);
        }

        while (code) {
                switch (code) {
                case netdev_code(transmit):   get_change(dev.transmit);    break;
                case netdev_code(receive):    get_change(dev.receive);    break;
                case netdev_code(collisions): get_change(dev.collisions);  break;
                case netdev_code(status):     get_object(dev.status);      break;
                }
                get_object(code);
        }
        netdev_update(host, device, &dev);

        return 0;
        
msg_error:
        return -1;
}

int add_monitor(struct host_info *host, int mtype)
{
        struct monitor_info *monitor;

        monitor = calloc(1, sizeof (struct monitor_info));

        if (!monitor) {
                syslog(LOG_WARNING, "cannot add monitor: not enough memory");
                return -ENOMEM;
        }

        host->monitor[mtype] = monitor;

        return 0;
}

void monitor_dispatch(void)
{
	struct in_addr addr;
	unsigned short port;
	char tag;
	short end;
	int mon_type;
	struct host_info *host;
	struct monitor_info *monitor = NULL;
	struct sockaddr_in mon_addr = {AF_INET};
	socklen_t addrlen;
	int err;

next_msg:
	addrlen = sizeof(mon_addr);
	msglen = recvfrom(monitor_socket, &monbuff, sizeof(struct monitor_msg),
			  MSG_NOSIGNAL, &mon_addr, &addrlen);
	if (msglen == -1)
		return;

	addr = mon_addr.sin_addr;
	port = ntohs(mon_addr.sin_port);
        mon_addr.sin_family = AF_INET;

#ifndef DEBUG
	if (port >= 1024)          /* Authorization check. Only root may bind */
		goto next_msg;     /* ports below 1024 */
#endif
        
        gettimeofday(&current_tv, NULL);
	msg_time = monbuff.time;

        host = hosts_by_addr[addr_to_id(addr)];
        if (!host || host->addr.s_addr != addr.s_addr)
                goto system_error;

        /*
         * This test is for integrated monitor. Its message can contain
         * data from multiple monitors.
         */
        mon_type = port - CISD_PORT;

        if (!host->monitor[mon_type]) {
                if (add_monitor(host, mon_type) == 0)
                        host->monitor[mon_type]->seqnum = monbuff.seqnum-1;
                else
                        goto next_msg;
        }

        monitor = host->monitor[mon_type];

	reply.code = 1;
	
        /* Wait for init message */
        if (!(monitor->flags&CONSISTENT) && !(monbuff.flags&MSG_INIT)) {
                reply.code = -EAGAIN;
                goto send_reply;
        }

	/* We loosed some messages. Ask for full message */
        if (++(monitor->seqnum) != monbuff.seqnum /*|| !monbuff.seqnum*/) {
		reply.code = -EFAULT;
		goto send_reply;
	}

        buffptr = (char *) monbuff.data;

        /* Parse the message */
        while (bufflen < msglen) {
		end = msglen;
		
                get_object(tag);
                get_object(end);

                if (msglen < end)
                        goto msg_error;

                mon_type = monitor_type(tag);
                if (!host->monitor[mon_type] && add_monitor(host, mon_type) != 0) {
                        buffptr = (char *) &monbuff + end;
                        continue;
                }
                monitor = host->monitor[mon_type];


                if (mon_type != PROCMON ||
                    (msg_type(tag) != MON_NEW && msg_type(tag) != MON_ASYNC_CHANGES)) {
                        monitor->local_time = msg_time;
                        monitor->time_tv    = current_tv;
                }

                if (!dispatch_fn[mon_type][msg_type(tag)])
                        goto msg_error;

                err = dispatch_fn[mon_type][msg_type(tag)](host, end);
                if (err)
                        goto msg_error;

		if (is_init_msg(tag))
                        goto send_reply;
	}
        goto next_msg;
        
msg_error:
send_reply:
        reply.seqnum = monitor->seqnum;
	sendto(monitor_socket, &reply, sizeof(reply), SOCKFLAGS, &mon_addr, addrlen);
        goto next_msg;

system_error:
	{
		int return_code = -EPERM;
		
		sendto(monitor_socket, &return_code, sizeof(return_code),
		       SOCKFLAGS, &mon_addr, addrlen);
	}
        goto next_msg;
}

void cleanup (int sig)
{
        (void) pmap_unset(prognum, CIS_VER);

#ifndef DEBUG
//        syslog(LOG_INFO, "terminated");
        closelog();
#endif
        exit(0);
}

int main (int argc,char *argv[])
{
        struct timeval t;
        
	register SVCXPRT *transp_tcp;
	int sock;
	struct sockaddr_in addr;
        int flags;
	fd_set rfds;
	int err;
	char myhostname[MAXHOSTNAMELEN];
	struct hostent *hent;

	get_myaddress(&srv_addr);
	if (gethostname(myhostname, MAXHOSTNAMELEN) == -1) {
		syslog(LOG_ERR, "cannot get my hostname");
		exit(1);
	}

	myaddr = srv_addr.sin_addr;

	if (myaddr.s_addr == 0x0100007f) {
		if (!(hent = gethostbyname(myhostname))) {
			syslog(LOG_ERR, "cannot get my address");
			exit(1);
		}
		myaddr = *(struct in_addr *) hent->h_addr;
	}
	
#ifndef DEBUG
        if (getuid()) {
                fprintf(stderr, "You don't have permission to run cis server.\n");
                exit (1);
        }

        openlog("cis", LOG_CONS, LOG_DAEMON);

        if (!getuid()) {
                /*
                 * Look at the portmapper whether there is a cis daemon already registered.
                 */

		if (pmap_getport(&srv_addr, CIS_PROG, CIS_VER, IPPROTO_TCP)) {
			syslog(LOG_ERR, "cis daemon is already registered");
                        exit(1);
		}
	}
#endif

	setlocale(LC_ALL,"");
	setlocale (LC_NUMERIC, "C");

	/*
         * Parse command line for parameters.
	 */
	parse_args(argc, argv);
	reply.interval = interval;

        parse_config();
	/*
	 * We need to unregister rpc service before terminating ...
	 */
	signal(SIGINT, cleanup);
	signal(SIGTERM, cleanup);
	signal(SIGPIPE, SIG_IGN);

	/*
	 * Initialize random number generator for authentication.
	 */
	gettimeofday(&t, NULL);
        srand((unsigned int) t.tv_usec);

	monitor_socket = socket(PF_INET, SOCK_DGRAM, 0);
	if (monitor_socket == -1) {
		syslog(LOG_ERR, "cannot create UDP server socket");
		exit(1);
        }

	addr.sin_family = AF_INET;
	addr.sin_port = htons(CISD_PORT);
        addr.sin_addr.s_addr = htonl (INADDR_ANY);
	err = bind (monitor_socket, (struct sockaddr *) &addr, sizeof (addr));
	if (err < 0) {
		syslog(LOG_ERR, "cannot bind UDP server socket");
		exit(1);
        }
	/*
	 * Enable nonblocking mode on socket.
	 */
	flags = fcntl(monitor_socket, F_GETFL, 0);
	if (flags == -1) {
		syslog(LOG_ERR, "cannot get UDP server socket flags");
		exit(1);
	}

	if (fcntl(monitor_socket, F_SETFL, (flags|FNDELAY)) == -1) {
		syslog(LOG_ERR, "cannot set UDP server socket flags");
		exit(1);
	};


#ifdef DEBUG
        (void) pmap_unset(CIS_PROG, CIS_VER);
#endif

        /*
         * Make a server point and register it with rpcbinder.
         */
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		syslog(LOG_ERR, "cannot create TCP server socket");
		exit(1);
	}


	addr.sin_family = AF_INET;
        addr.sin_port = htons(CISD_PORT-1);
        addr.sin_addr.s_addr = htonl (INADDR_ANY);
        err = bind (sock, (struct sockaddr *) &addr, sizeof (addr));
        if (err < 0) {
                syslog(LOG_ERR, "cannot bind TCP server socket");
                exit(1);
        }
	
        transp_tcp = svctcp_create(sock, 0, 0);
        if (transp_tcp == NULL) {
                syslog(LOG_ERR, "cannot create tcp service");
                exit(1);
        }

        if (!svc_register(transp_tcp, prognum, CIS_VER, dispatcher, IPPROTO_TCP)) {
                syslog(LOG_ERR, "cannot register tcp service with the rpcbinder");
                exit(1);
        }

#ifndef DEBUG
        /*
         * It's time to do the magic to become a daemon.
         */
        daemonize();
#endif

        syslog(LOG_INFO, "started");

        for(;;) {

	        rfds = svc_fdset;
                FD_SET(monitor_socket, &rfds);
                
                switch (select (FD_SETSIZE, &rfds, NULL, NULL, NULL)) {

                case -1:
                        continue;

                default:
			if (FD_ISSET(monitor_socket, &rfds))
				monitor_dispatch();

                        FD_CLR(monitor_socket, &rfds);
                        svc_getreqset (&rfds);
                }
	}
}
