/*
 * Cluster Information Service - a monitoring system for Linux clusters
 * Copyright (C) 2000 Institute of Informatics, Slovak Academy of Sciences.
 * Written by Jan Astalos (astalos.ui@savba.sk)
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * CIS server daemon - parsing of configuration file.
 */

#include "cis.h"
#include "cisd.h"

extern struct cis_info my_info;
extern struct in_addr myaddr;
extern char *config_file;
extern int nhosts;
extern struct host_info **hosttab, *hosts_by_addr[256];
extern char **perf_types;
extern int nperf_types;

char *commands[]   = {"#", "Name", "PerfTypes", "PerfValues", "Reliability"};
char *rel_types[]  = {"minimal", "low", "average", "high", "maximal"};
char *yesno[]      = {"no", "yes"};

enum {
        CMD_COMMENT,
        CMD_NAME,
        CMD_PERFTYPES,
        CMD_PERFVALUES,
	CMD_RELIABILITY,
	CMD_TABLEN,
};

int parse_cmd (char *cmds[], int n, char **c)
{
	int i;

	*c += strspn (*c, " \t");

	for (i = 0; i < n; i++)
		if (!strncasecmp (*c, cmds[i], strlen (cmds[i]))) {
			*c += strlen(cmds[i]);
			return (i);
		}
	return -1;
}

int parse_config(void)
{
	struct host_info *host;
        struct hostent *hent;
	FILE *file = NULL;
	int line = 0;
	char name[MAXHOSTNAMELEN];
	char *c, *type;
	int i, cmd;
	int rel = 0;
        float *performance = NULL;
        struct in_addr addr;
        char namebuf[64];
        
	int buflen = 0;
	char *buf  = NULL;

	/*
         * Open the config file.
         */
	if (!(file = fopen(config_file ? config_file : DEF_CONFIG, "r"))) {
		syslog(LOG_ERR, "cannot open config file %s", config_file);
		exit (1);
	}

        nhosts = 0;
        memset (&hosts_by_addr, 0, sizeof (hosts_by_addr));
        hosttab = NULL;
	
	while (getline(&buf, &buflen, file) != -1) {

		line++;
		c = buf;
		
		cmd = parse_cmd(commands, CMD_TABLEN, &c);
		if (cmd == 0) continue;

		c += strspn(c, " \t");       /* strip white spaces */

                switch (cmd) {
                        
		case CMD_NAME:
                        sscanf (c, "%s", namebuf);
                        my_info.name = strdup(namebuf);
                        break;
			
		case CMD_PERFTYPES:

			c = strdup(c);
			type = strtok(c, ", \t\n");

			do {
				perf_types = realloc(perf_types, (nperf_types+1)* sizeof(char *));
				if (!perf_types)
					goto nomem;

				perf_types[nperf_types++] = type;

			} while ((type = strtok(NULL, ", \t\n")));

			performance = calloc(nperf_types, sizeof(float));

			break;

		case CMD_PERFVALUES:

			for (i = 0; *c; i++, c++) {

				performance[i] = atof(c);

				c = strchr(c, ',');
				if (i == nperf_types || !c)
					break;

				c++;
			}
			break;

		case CMD_RELIABILITY:

			rel = parse_cmd(rel_types, REL_TABLEN, &c);
			if (rel == -1)
				goto error;

			break;

		default:

			if (sscanf(c, "%s", name) != 1 || *name == '\n')
				continue;

			if (++nhosts == 255) {
                                syslog(LOG_ERR, "too much hosts");
                                exit(1);
                        }

                        if ((hent = gethostbyname(name)) == NULL) {
				syslog(LOG_ERR, "cannot get the address of host %s", name);
				exit(1);
			}

                        addr = *(struct in_addr *) hent->h_addr;
                        
                        if (hosts_by_addr[addr_to_id(addr)]) {
                                syslog(LOG_ERR, "ambiguous hosts %s %s (the same last byte of address)",
                                       name, hosts_by_addr[addr_to_id(addr)]->hostinfo.name);
                                exit(1);
                        }

			host = calloc(1, sizeof (struct host_info));
			if (!host)
				goto nomem;

/*                        if (*my_info.name)
                                snprintf(host->hostinfo.name, HOST_NAME_LEN, "%s.%s", my_info.name, name);
                        else*/
			snprintf(host->hostinfo.name, HOST_NAME_LEN, "%s", name);

                        host->addr = addr;

			host->hostinfo.reliability = rel;
			host->performance = malloc(nperf_types * sizeof(float));
			if (!host->performance)
				goto nomem;

                        memcpy (host->performance, performance, nperf_types*sizeof(float));

                        host->hostinfo.addr.server_addr = myaddr;
                        host->hostinfo.addr.node_id     = nhosts;

                        hosttab = realloc (hosttab, (nhosts+1)*sizeof(struct host_info *));
                        if (!hosttab)
                                goto nomem;

                        hosttab[nhosts] = host;
                        hosts_by_addr[addr_to_id(addr)] = host;
                }
	}
	fclose (file);

       	return 1;
nomem:
	syslog(LOG_ERR, "not enough memory");
       	exit(1);

error:
	syslog(LOG_ERR, "bad entry in config file at line %d", line);
	exit (1);
}

