/*
 * Cluster Information Service - a monitoring system for Linux clusters
 * Copyright (C) 2000 Institute of Informatics, Slovak Academy of Sciences.
 * Written by Jan Astalos (astalos.ui@savba.sk)
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * CIS server daemon - monitor message dispatch routines
 */

#include <errno.h>

#include "cis.h"
#include "cisd.h"

extern float interval;
extern struct timeval current_tv;
extern struct monitor_msg monbuff;
extern char *buffptr;
#define bufflen (buffptr - (char *) &monbuff)
extern unsigned long msg_time;

static void free_monitor(struct host_info *host, int mon_type)
{
        free(host->monitor[mon_type]);
        host->monitor[mon_type] = NULL;
}

static void init_monitor(struct monitor_info *monitor, unsigned long start, unsigned short ct)
{
        monitor->clk_interval = interval*ct;
        monitor->start_tv     = current_tv;
        monitor->start_clk    = start;
        monitor->seqnum       = monbuff.seqnum;
        monitor->flags        = CONSISTENT;
        monitor->ct_per_sec   = ct;
}

static int cismon_init(struct host_info *host, int end)
{
        struct monitor_info *monitor = host->monitor[MONITOR];
        unsigned long  start_clk;
        int ct_per_sec;

        get_object(start_clk);
        get_object(ct_per_sec);

	init_monitor(monitor, start_clk, ct_per_sec);
	
	host->hostinfo.ct_per_sec = ct_per_sec;

	return 0;

msg_error:
        return -1;
}

static int cismon_exit(struct host_info *host, int end)
{
        free_monitor(host, MONITOR);
        return 0;
}

///////////////////////////////////////////////////////////////////////////////
int sys_change(struct cis_hostinfo *hi, int end);
        
static int sysmon_init(struct host_info *host, int end)
{
        struct monitor_info *monitor = host->monitor[SYSMON];
        unsigned long  start_clk;
        struct utsname uname;
        int ct_per_sec;
        unsigned long totalram, totalswap;
        struct in_addr addrtab[MAX_ADDR_NUM];

        get_object(start_clk);
        get_object(ct_per_sec);

        get_object(uname);
        get_object(totalram);
        get_object(totalswap);
        get_object(addrtab);

        init_monitor(monitor, start_clk, ct_per_sec);

        host->hostinfo.totalram   = totalram;
        host->hostinfo.totalswap  = totalswap;
        host->hostinfo.ct_per_sec = ct_per_sec;
	host->uname               = uname;

        memcpy(&host->hostinfo.addrtab, &addrtab, MAX_ADDR_NUM*sizeof(struct in_addr));

/*        if(*cis_name)
                snprintf(host->hostinfo.name, CIS_HOSTNAMELEN, "%s.%s", cis_name, uname.nodename);
        else
                snprintf(host->hostinfo.name, CIS_HOSTNAMELEN, "%s", uname.nodename);
*/
        init_history(&host->ctx_swtch_history, monbuff.time);
        init_history(&host->swpin_history, monbuff.time);
        init_history(&host->swpout_history, monbuff.time);
        init_history(&host->disks_read_history, monbuff.time);
        init_history(&host->disks_write_history, monbuff.time);

        return 0;

msg_error:
        return -1;
}

static int sysmon_exit(struct host_info *host, int end)
{
        free_monitor(host, SYSMON);
        return 0;
}

static int sysmon_msg(struct host_info *host, int end, int full)
{
        struct cis_hostinfo newhi = host->hostinfo, *hi = &host->hostinfo;
        struct monitor_info *monitor = host->monitor[SYSMON];
        
        monitor->flags |= CONSISTENT;

        if (full) {
                get(&newhi, HOSTINFO_LEN);
        } else if (sys_change(&newhi, end))
                goto msg_error;

        if (newhi.idle != hi->idle)
                host->idle_at = msg_time;
        if (newhi.ctx_swtch != hi->ctx_swtch)
                history_add(&host->ctx_swtch_history,
                            monitor->local_time,
                            newhi.ctx_swtch - hi->ctx_swtch);

        if (newhi.swpin != hi->swpin)
                history_add(&host->swpin_history,
                            monitor->local_time,
                            newhi.swpin - hi->swpin);
        
        if (newhi.swpout != hi->swpout)
                history_add(&host->swpout_history,
                            monitor->local_time,
                            newhi.swpout - hi->swpout);
        
	if (newhi.disks_read != hi->disks_read)
                history_add(&host->disks_read_history,
                            monitor->local_time,
                            newhi.disks_read - hi->disks_read);

	if (newhi.disks_write != hi->disks_write)
                history_add(&host->disks_write_history,
                            monitor->local_time,
			    newhi.disks_write - hi->disks_write);
	
	*hi = newhi;
	
        return 0;

msg_error:
        return -1;
}

static int sysmon_full(struct host_info *host, int end)
{
        return sysmon_msg(host, end, 1);
}

static int sysmon_changes(struct host_info *host, int end)
{
        return sysmon_msg(host, end, 0);
}

///////////////////////////////////////////////////////////////////////////////
void proc_test(struct host_info *host, struct cis_procinfo *p);
void proc_remove(struct host_info *host, struct process_info *process);


void clear_proclist(struct host_info *host)
{
        while (host->proclist)
		proc_remove(host, host->proclist);
}

static int procmon_init(struct host_info *host, int end)
{
        struct monitor_info *monitor = host->monitor[PROCMON];
        unsigned long  start_clk;
        int ct_per_sec;

        get_object(start_clk);
        get_object(ct_per_sec);

        init_monitor(monitor, start_clk, ct_per_sec);

        if (host->proc_ht)
                clear_proclist(host);
        else host->proc_ht = hash_create(PROC_HT_LENGTH,
                                         OFFSET(process_info, data.pid),
                                         sizeof (int), 0);
        if (!host->proc_ht){
                free_monitor(host, PROCMON);
                return -ENOMEM;
        }

        host->hostinfo.ct_per_sec = ct_per_sec;
        host->slsum = 0;

        return 0;

msg_error:
        return -1;
}

static int procmon_exit(struct host_info *host, int end)
{
        clear_proclist(host);
        free_monitor(host, PROCMON);
        return 0;
}

int proc_change(struct host_info *host, unsigned int pid, int end);


static int procmon_full(struct host_info *host, int end)
{
        struct monitor_info *monitor = host->monitor[PROCMON];
        struct process_info *process, *p;
	struct cis_procinfo proc;

        monitor->flags |= CONSISTENT;

        if (monbuff.flags & MSG_FIRST)
                for (p = host->proclist; p; p = p->next)
                        p->flags &= ~CONSISTENT;

        while (bufflen < end) {
                get(&proc, PROCINFO_LEN);
                proc_test(host, &proc);
        }

        if (!(monbuff.flags & MSG_LAST))
                return 0;

        /*
         * Free all terminated processes.
         */
        p = host->proclist;
        while (p) {
                process = p;
                p = p->next;
                if (process->flags&CONSISTENT) continue;

                proc_remove(host, process);
        }

        return 0;
        
msg_error:
        return -1;
}

static int procmon_changes(struct host_info *host, int end)
{
        unsigned int  pid;
        
        while (bufflen < end) {
                get_object(pid);
                if (proc_change(host, pid, end))
                        goto msg_error;
        }
        return 0;

msg_error:
        return -1;
}

static int procmon_new(struct host_info *host, int end)
{
        struct cis_procinfo proc;

        get(&proc, PROCINFO_LEN);
        proc_test(host, &proc);
        return 0;

msg_error:
        return -1;
}

///////////////////////////////////////////////////////////////////////////////
void sock_test(struct host_info *host, struct cis_sockinfo *s);
void sock_remove (struct host_info *host, struct socket_info *s);
int sock_change(struct host_info *host, struct cis_sockinfo *s, int end);

void clear_socklist(struct host_info *host)
{
	struct socket_info *socket;

	while (host->socklist) {
		socket = host->socklist;
		host->socklist = host->socklist->next;
		hash_remove(host->sock_ht, &socket->data);
		free(socket);
        }
}

void clear_sockets(struct host_info *host, unsigned long time)
{
	struct socket_info *s, *tmp;

	time -= SOCKET_TIMEOUT * host->hostinfo.ct_per_sec;

	for (s = host->socklist; s;) {
		tmp = NULL;

		if (!(s->flags&CONSISTENT) &&
                    s->sent_history.last->time < time &&
		    s->rcvd_history.last->time < time)
			tmp = s;

		s = s->next;

                if (tmp) {
			list_remove(host->socklist, tmp);
			free(tmp);
		}
	}
}

static int sockmon_init(struct host_info *host, int end)
{
        struct monitor_info *monitor = host->monitor[SOCKMON];
        unsigned long start_clk;
        int ct_per_sec;

        get_object(start_clk);
        get_object(ct_per_sec);

        init_monitor(monitor, start_clk, ct_per_sec);

        if (host->sock_ht) {
                clear_socklist(host);
                return 0;
        }
        host->sock_ht = hash_create(SOCK_HT_LENGTH,
                                    OFFSET(socket_info, data),
                                    SOCKINFO_HDRLEN,
                                    OFFSET(cis_sockinfo, sport));
        if (!host->sock_ht) {
                free_monitor(host, SOCKMON);
                return -ENOMEM;
        }

        host->hostinfo.ct_per_sec = ct_per_sec;

        return 0;

msg_error:
        return -1;
}

static int sockmon_exit(struct host_info *host, int end)
{
        clear_socklist(host);
        free_monitor(host, SOCKMON);
        return 0;
}

static int sockmon_full(struct host_info *host, int end)
{
        struct cis_sockinfo sock;
        struct socket_info *s;
        
        if (monbuff.flags & MSG_FIRST)
                for (s = host->socklist; s; s = s->next)
                        s->flags &= ~CONSISTENT;
        
        while (bufflen < end) {
                get(&sock, SOCKINFO_LEN);
                sock_test(host, &sock);
        }
        
        if (!(monbuff.flags & MSG_LAST))
                return 0;
        /*
         * Free all closed sockets.
         */
        clear_sockets(host, msg_time);
        
        for (s = host->socklist; s; s = s->next) {
                if (s->flags&CONSISTENT) continue;
                sock_remove(host, s);
        }

        return 0;
        
msg_error:
        return -1;
}

static int sockmon_changes(struct host_info *host, int end)
{
        struct cis_sockinfo sock;

        while (bufflen < end) {
                get(&sock, SOCKINFO_HDRLEN);
                if (sock_change(host, &sock, end))
                        goto msg_error;
        }

        clear_sockets(host, msg_time);

        return 0;

msg_error:
        return -1;
}

static int sockmon_new(struct host_info *host, int end)
{
        struct cis_sockinfo sock;

        get(&sock, SOCKINFO_LEN);
        sock_test(host, &sock);

        return 0;

msg_error:
        return -1;
}

///////////////////////////////////////////////////////////////////////////////
int netdev_test(struct host_info *host, struct cis_netdevinfo *d);
int netdev_change(struct host_info *host, unsigned char id, int end);

void clear_netdevlist(struct host_info *host)
{
	struct netdev_info *device;
        
        while (host->netdevlist) {
		device = host->netdevlist;
		host->netdevlist = host->netdevlist->next;
		free(device);
	}
}

static int netmon_init(struct host_info *host, int end)
{
        struct monitor_info *monitor = host->monitor[NETMON];
        unsigned long start_clk;
        int ct_per_sec;

        get_object(start_clk);
        get_object(ct_per_sec);

        init_monitor(monitor, start_clk, ct_per_sec);

        clear_netdevlist(host);

        host->hostinfo.ct_per_sec = ct_per_sec;

        return 0;

msg_error:
        return -1;
}

static int netmon_exit(struct host_info *host, int end)
{
        clear_netdevlist(host);
        free_monitor(host, NETMON);
        return 0;
}

static int netmon_full(struct host_info *host, int end)
{
	struct cis_netdevinfo dev;
        struct netdev_info *d;
        
        if (monbuff.flags & MSG_FIRST) {
                for (d = host->netdevlist; d; d = d->next)
                        d->flags &= ~CONSISTENT;
        }
        
        while (bufflen < end) {
                get(&dev, NETDEVINFO_LEN);
                netdev_test(host, &dev);
        }
        
        if (!(monbuff.flags & MSG_LAST))
                return 0;
        
        /*
         * Clear status for stopped devices.
         */
        for (d = host->netdevlist; d; d = d->next) {
                if (d->flags&CONSISTENT) continue;
                d->data.status = 0;
        }

        return 0;
        
msg_error:
        return -1;
}

static int netmon_changes(struct host_info *host, int end)
{
        unsigned char id;

        while (bufflen < end) {
                get_object(id);
                if (netdev_change(host, id, end))
                        goto msg_error;
        }

        return 0;

msg_error:
        return -1;
}

static int netmon_new(struct host_info *host, int end)
{
	struct cis_netdevinfo dev;

        get(&dev, NETDEVINFO_LEN);
        netdev_test(host, &dev);

        return 0;

msg_error:
        return -1;
}

///////////////////////////////////////////////////////////////////////

static int lmmon_init(struct host_info *host, int end)
{
        struct monitor_info *monitor = host->monitor[LMMON];
        unsigned long start_clk;
        int ct_per_sec;
        unsigned char nsensors;
        struct cis_lmsinfo *newtab;

        get_object(start_clk);
        get_object(ct_per_sec);

        get_object(nsensors);

        init_monitor(monitor, start_clk, ct_per_sec);

        newtab = realloc(host->lmstab, nsensors * sizeof(struct cis_lmsinfo));
        if (!newtab) {
                free_monitor(host, LMMON);
                return -ENOMEM;
        }
        host->lmstab = newtab;
        host->lmstab_len = nsensors;
        memset(host->lmstab, 0, nsensors * sizeof(struct cis_lmsinfo));

        host->hostinfo.ct_per_sec = ct_per_sec;

        return 0;

msg_error:
        return -1;
}

static int lmmon_exit(struct host_info *host, int end)
{
        free(host->lmstab);
        host->lmstab = NULL;
        host->lmstab_len = 0;
        free_monitor(host, LMMON);
        return 0;
}

static int lmmon_full(struct host_info *host, int end)
{
        unsigned char id;

        while (bufflen < end) {
                get_object(id);
                if (id >= host->lmstab_len)
                        continue;
                get_object(host->lmstab[(int) id]);
        }

        return 0;
        
msg_error:
        return -1;
}

static int lmmon_changes(struct host_info *host, int end)
{
        unsigned char id;

        while (bufflen < end) {
                get_object(id);
                if (id >= host->lmstab_len)
                        continue;
                get_object(host->lmstab[(int) id].value);
        }

        return 0;

msg_error:
        return -1;
}

int (*dispatch_fn[NMONITORS][NMSGTYPES])(struct host_info *, int) = {
	{}, /* unused */
	{cismon_init,  cismon_exit}, /* cismon */
        {sysmon_init,  sysmon_exit,  sysmon_full,  sysmon_changes},
        {procmon_init, procmon_exit, procmon_full, procmon_changes, procmon_new, procmon_changes},
        {sockmon_init, sockmon_exit, sockmon_full, sockmon_changes, sockmon_new},
        {netmon_init,  netmon_exit,  netmon_full,  netmon_changes,  netmon_new},
        {lmmon_init,   lmmon_exit,   lmmon_full,   lmmon_changes},
};

