/*
 * Cluster Information Service - a monitoring system for Linux clusters
 * Copyright (C) 2000 Institute of Informatics, Slovak Academy of Sciences.
 * Written by Jan Astalos (astalos.ui@savba.sk)
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * CIS server daemon - client RPC calls dispatch routines.
 */

#include "cis.h"
#include "cis_xdr.h"
#include "cisd.h"
#include "cis_clnt.h"

extern bool_t xdr_process_common(XDR *handle, struct cis_procinfo *p);
extern struct host_info **hosttab, *hosts_by_addr[256];
extern int nhosts;

/*
 * Output filters
 */
/* For messages from clients */
extern struct timeval current_tv;
extern char **perf_types;
extern int nperf_types;
extern struct cis_info my_info;

struct host_info *find_host(struct in_addr addr);

int count_pCPU(struct host_info *host, struct process_info *p, unsigned long current_time);
int availableCPU(struct host_info *host);
int monitor_idle(struct monitor_info *monitor);
unsigned long count_rate(struct history *h, unsigned long time, unsigned long interval, unsigned long cticks);

/*bool_t xdr_list_req(XDR *handle, struct clnt_req *req)
{
        if (!xdr_float(handle, &req->interval))
                return (FALSE);
	if (!xdr_u_short(handle, &req->uid))
                return (FALSE);

        return (TRUE);
} */

int perf_index (char *perf)
{
	int i;

	for (i = 1; i < nperf_types; i++)
		if (!strcmp(perf_types[i], perf))
			return i;
	return 0;
}

struct host_info *find_host_by_name(char *name)
{
        int i;
        
        for(i = 1; i < nhosts; i++)
                if (hosttab[i] && !(strcmp(hosttab[i]->hostinfo.name, name)))
                        return hosttab[i];
        return NULL;
}

bool_t xdr_host_info(XDR *handle, struct host_req *req)
{
        struct host_info *host;
        struct cis_hostinfo tmp;
	char status = HOST_AVAILABLE;
        unsigned long interval, ctps, time;

        host = req->id ? host = hosttab[req->id] : find_host_by_name(req->name);
        if (!host)
                status = HOST_UNKNOWN;
        else if (!host->monitor[SYSMON])
                status = HOST_NOT_AVAILABLE;
        else if (monitor_idle (host->monitor[SYSMON]))
                status = HOST_NOT_AVAILABLE;

        if (!xdr_char(handle, &status))
                return (FALSE);

        if (status == HOST_UNKNOWN)
                return TRUE;

        tmp = host->hostinfo;

        tmp.performance = host->performance[perf_index(req->perf_type)];
        if (!xdr_static_hostinfo(handle, &tmp))
                return (FALSE);

        if (status != HOST_AVAILABLE)
                return TRUE;

        tmp.CPU_available = availableCPU(host);

        ctps = tmp.ct_per_sec;
        interval = req->interval * ctps;
        
        time = monitor_time(host->monitor[SYSMON]);
        tmp.ctx_swtch   = count_rate(&host->ctx_swtch_history,   time, interval, ctps);
        tmp.swpin       = count_rate(&host->swpin_history,       time, interval, ctps);
        tmp.swpout      = count_rate(&host->swpout_history,      time, interval, ctps);
        tmp.disks_read  = count_rate(&host->disks_read_history,  time, interval, ctps);
        tmp.disks_write = count_rate(&host->disks_write_history, time, interval, ctps);

        if (!xdr_dynamic_hostinfo(handle, &tmp))
                return (FALSE);

        return (TRUE);
}

bool_t xdr_hostlist(XDR *handle, struct host_req *req)
{
        int i, n = 0;

        for(i = 1; i <= nhosts; i++)
                if (hosttab[i]->hostinfo.reliability >= req->reliability)
                        n++;

        if (!xdr_int(handle, &n))
                return (FALSE);

        if (!n)
		return (TRUE);

        for(i = 1; i <= nhosts; i++)
                if (hosttab[i]->hostinfo.reliability >= req->reliability) {
                        req->id = hosttab[i]->hostinfo.addr.node_id;

                        if (!xdr_host_info(handle, req))
                                return (FALSE);
                }

	return (TRUE);
}


bool_t xdr_host_update(XDR *handle, struct host_update_req *req)
{
        struct host_info *host;
	struct cis_hostinfo *hi;
        char stat;
	unsigned long time, interval, ctps, rate;
        int i, code = req->update_code;
	
        for (i = 0; i < req->length; i++) {
                stat = HOST_AVAILABLE;
                if (!(host = hosttab[req->idtab[i]]))
                        stat = HOST_UNKNOWN;
		else if (!host->monitor[SYSMON])
			stat = HOST_NOT_AVAILABLE;
		else if (monitor_idle (host->monitor[SYSMON]))
			stat = HOST_NOT_AVAILABLE;

                if (!xdr_char(handle, &stat))
                        return (FALSE);

                if (stat != HOST_AVAILABLE)
                        continue;

                ctps = host->hostinfo.ct_per_sec;
                interval = req->interval * ctps;
                time = monitor_time(host->monitor[SYSMON]);
                hi = &host->hostinfo;

                if (code & CIS_PROCNUM)
                        if (!xdr_u_short(handle, &hi->procnum))
                                return (FALSE);

                if (code & CIS_LOADS)
                        if (!xdr_vector(handle, (char *) hi->loads, 3, sizeof(short), (xdrproc_t) xdr_u_short))
                                return (FALSE);

                if (code & CIS_RAM) {
                        if (!xdr_u_long(handle, &hi->totalram))
                                return (FALSE);
                        if (!xdr_u_long(handle, &hi->freeram))
                                return (FALSE);
                        if (!xdr_u_long(handle, &hi->sharedram))
                                return (FALSE);
                        if (!xdr_u_long(handle, &hi->bufferram))
                                return (FALSE);
                        if (!xdr_u_long(handle, &hi->cachedram))
                                return (FALSE);
                }

                if (code & CIS_SWAP) {
                        if (!xdr_u_long(handle, &hi->totalswap))
                                return (FALSE);
                        if (!xdr_u_long(handle, &hi->freeswap))
                                return (FALSE);
                        rate = count_rate(&host->swpin_history, time, interval, ctps);
                        if (!xdr_u_long(handle, &rate))
				return (FALSE);
                        rate = count_rate(&host->swpout_history, time, interval, ctps);
			if (!xdr_u_long(handle, &rate))
				return (FALSE);
                }
                
		if (code & CIS_CPU) {
                        hi->CPU_available = availableCPU(host);
                        if (!xdr_u_short(handle, &hi->CPU_available))
                                return (FALSE);
                }

		if (code & CIS_IDLE)
			if (!xdr_u_long(handle, &hi->idle))
				return (FALSE);

		if (code & CIS_CTXSWTCH) {
			rate = count_rate(&host->ctx_swtch_history, time, interval, ctps);
			if (!xdr_u_long(handle, &rate))
				return (FALSE);
		}
		if (code & CIS_DISK) {
                        rate = count_rate(&host->disks_read_history, time, interval, ctps);
                        if (!xdr_u_long(handle, &rate))
                                return (FALSE);

                        rate = count_rate(&host->disks_write_history, time, interval, ctps);
                        if (!xdr_u_long(handle, &rate))
                                return (FALSE);
                }
	}

        return (TRUE);
}

#define user_match(a,b) (a == 0xffff || (a ? (a == b) : (b > MIN_USER_UID)))

bool_t xdr_proclist_rep(XDR *handle, struct list_req *req)
{
        struct host_info *host = hosttab[req->id];
        struct process_info *p;
        unsigned short uid = req->uid;
        unsigned long time, interval, ctps;
        int n, err = 0;

        if (!host)
                err = -HOST_UNKNOWN;
        else if (!host->monitor[PROCMON] || monitor_idle(host->monitor[PROCMON]))
                err = -HOST_NOT_AVAILABLE;

        if (err) {
                if (!xdr_int(handle, &err))
                        return (FALSE);
                return TRUE;
        }

        time = monitor_time(host->monitor[PROCMON]);
        ctps = host->hostinfo.ct_per_sec;
        interval = req->interval * ctps;
        
        for (p = host->proclist, n = 0; p; p = p->next) {
                if (!user_match(uid, p->data.uid))
                        continue;
                p->data.pCPU = count_pCPU(host, p, time);
                n++;
        }

        if (!xdr_int(handle, &n))
                return (FALSE);

	if (!n)
                return (TRUE);

	for (p = host->proclist; p; p = p->next) {
                if (!user_match(uid, p->data.uid))
			continue;

		if (!xdr_process_common(handle, &p->data))
                        return (FALSE);

                p->majflt_rate = count_rate(&p->majflt_history, time, interval, ctps);
		p->read_rate   = count_rate(&p->read_history,   time, interval, ctps);
		p->write_rate  = count_rate(&p->write_history,  time, interval, ctps);

		if (!xdr_u_long(handle, &p->GMT_start_time))
                        return (FALSE);
                if (!xdr_u_long(handle, &p->stime_sec))
                        return (FALSE);
                if (!xdr_u_long(handle, &p->utime_sec))
                        return (FALSE);
		if (!xdr_short(handle, &p->data.pCPU))
			return (FALSE);

		if (!xdr_u_long(handle, &p->majflt_rate))
			return (FALSE);
		if (!xdr_u_long(handle, &p->read_rate))
			return (FALSE);
		if (!xdr_u_long(handle, &p->write_rate))
			return (FALSE);
        }
        return (TRUE);
}

bool_t xdr_socklist_rep(XDR *handle, struct list_req *req)
{
        struct host_info *host = hosttab[req->id];
        struct socket_info *s;
	unsigned short uid = req->uid;
        unsigned long time, interval, ctps;
        int n, err = 0;

        if (!host)
                err = -HOST_UNKNOWN;
        else if (!host->monitor[SOCKMON] || monitor_idle(host->monitor[SOCKMON]))
                err = -HOST_NOT_AVAILABLE;

        if (err) {
                if (!xdr_int(handle, &err))
                        return (FALSE);
                return TRUE;
        }

        time = monitor_time(host->monitor[SOCKMON]);
        ctps = host->hostinfo.ct_per_sec;
        interval = req->interval * ctps;

	for (s = host->socklist, n = 0; s; s = s->next) {
		if (!user_match(uid, s->data.uid))
			continue;

		s->send_rate = count_rate(&s->sent_history, time, interval, ctps);
		s->recv_rate = count_rate(&s->rcvd_history, time, interval, ctps);

		if (!s->send_rate && !s->recv_rate)
                        continue;
		n++;
        }
        if (!xdr_int(handle, &n))
                return (FALSE);

        if (!n)
                return (TRUE);

        for (s = host->socklist; s; s = s->next) {
		if (!user_match(uid, s->data.uid))
			continue;

		if (!s->send_rate && !s->recv_rate)
                        continue;
                
                if (!xdr_socket_desc(handle, &s->data))
                        return (FALSE);

		if (!xdr_u_int(handle, &s->data.pid))
                        return (FALSE);
		if (!xdr_u_short(handle, &s->data.uid))
                        return (FALSE);

                if (!xdr_u_long(handle, &s->send_rate))
                        return (FALSE);
		if (!xdr_u_long(handle, &s->recv_rate))
                        return (FALSE);
        }
        return (TRUE);
}

bool_t xdr_netdevlist_rep(XDR *handle, struct list_req *req)
{
        struct host_info *host = hosttab[req->id];
        struct netdev_info *d;
        unsigned long time, interval, ctps;
        int n, err = 0;

        if (!host)
                err = -HOST_UNKNOWN;
        else if (!host->monitor[NETMON] || monitor_idle(host->monitor[NETMON]))
                err = -HOST_NOT_AVAILABLE;

        if (err) {
                if (!xdr_int(handle, &err))
                        return (FALSE);
                return TRUE;
        }

        time = monitor_time(host->monitor[NETMON]);
        ctps = host->hostinfo.ct_per_sec;
        interval = req->interval * ctps;

        for (d = host->netdevlist, n = 0; d; d = d->next, n++)
                ;

	if (!xdr_int(handle, &n))
                return (FALSE);

        if (!n)
                return (TRUE);

        for (d = host->netdevlist; d; d = d->next) {

		if (d->data.status) {
                        d->rxbytes_rate    = count_rate(&d->rxbytes_history, time, interval, ctps);
                        d->txbytes_rate    = count_rate(&d->txbytes_history, time, interval, ctps);
			d->collisions_rate = count_rate(&d->collisions_history, time, interval, ctps);
                } else
			d->rxbytes_rate = d->txbytes_rate = d->collisions_rate = 0;

		if (!xdr_netdev_desc(handle, &d->data))
                        return (FALSE);

                if (!xdr_u_long(handle, &d->rxbytes_rate))
                        return (FALSE);
                if (!xdr_u_long(handle, &d->txbytes_rate))
                        return (FALSE);
		if (!xdr_u_long(handle, &d->collisions_rate))
                        return (FALSE);
        }
        return (TRUE);
}
bool_t xdr_lmsensors(XDR *handle, struct list_req *req)
{
        struct host_info *host = hosttab[req->id];
	struct cis_lmsinfo *s;
        int n = 0, err = 0;

        if (!host)
                err = -HOST_UNKNOWN;
        else if (!host->monitor[LMMON] || monitor_idle(host->monitor[LMMON]))
                err = -HOST_NOT_AVAILABLE;

        if (err) {
                if (!xdr_int(handle, &err))
                        return (FALSE);
                return TRUE;
        }

        for (s = host->lmstab; s < host->lmstab + host->lmstab_len; s++)
		if (s->divider)
                        n++;

	if (!xdr_int(handle, &n))
                return (FALSE);

        for (s = host->lmstab; s < host->lmstab + host->lmstab_len; s++)
                if (s->divider)
                        if (!xdr_lmsinfo(handle, s))
                                return (FALSE);
        return (TRUE);
}
	

void clean_req(struct clnt_req *req)
{
	if (req->idtab)
		free(req->idtab);
	if (req->perf_type)
		free(req->perf_type);
	memset(req, 0, sizeof (struct clnt_req));
}


void dispatcher(struct svc_req *rqstp, SVCXPRT *transp)
{
        struct host_req hreq = {NULL, 0, 0., NULL, 0};
        struct host_update_req hureq = {0, 0., 0, NULL};
	struct list_req lreq;

        gettimeofday(&current_tv, NULL);

        /*
         * Get arguments
         */
        switch (rqstp->rq_proc) {

        case HOST_INFO:
        case HOST_LIST:
                if (!svc_getargs(transp, (xdrproc_t) xdr_host_req, (caddr_t) &hreq))
                        goto decode_error;
                break;

        case HOST_INFO_UPDATE:
                if (!svc_getargs(transp, (xdrproc_t) xdr_host_update_req, (caddr_t) &hureq))
                        goto decode_error;
                break;

        case PROCESS_LIST:
        case SOCKET_LIST:
        case NETDEV_LIST:
        case LM_SENSORS:
                if (!svc_getargs(transp, (xdrproc_t) xdr_list_req, (caddr_t) &lreq))
                        goto decode_error;
                break;
        }

        /*
         * Call dispatch routine
         */
        switch (rqstp->rq_proc) {

        case CIS_INFO:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_cisinfo, (caddr_t) &my_info))
                        goto system_error;
                break;

        case HOST_LIST:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_hostlist, (caddr_t) &hreq))
                        goto system_error;
                break;

        case HOST_INFO:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_host_info, (caddr_t) &hreq))
                        goto system_error;
                break;
                
        case HOST_INFO_UPDATE:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_host_update, (caddr_t) &hureq))
                        goto system_error;
                break;

        case PROCESS_LIST:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_proclist_rep, (caddr_t) &lreq))
                        goto system_error;
                break;
                
        case SOCKET_LIST:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_socklist_rep, (caddr_t) &lreq))
                        goto system_error;
                break;
                
        case NETDEV_LIST:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_netdevlist_rep, (caddr_t) &lreq))
                        goto system_error;
                break;

        case LM_SENSORS:
                if (!svc_sendreply(transp, (xdrproc_t) xdr_lmsensors, (caddr_t) &lreq))
                        goto system_error;
                break;

        default:
                svcerr_noproc(transp);
                return;
        }

        free(hreq.name);
        free(hreq.perf_type);
        free(hureq.idtab);

        return;

system_error:

        free(hreq.name);
        free(hreq.perf_type);
        free(hureq.idtab);

        svcerr_systemerr(transp);
        return;

decode_error:
	svcerr_decode(transp);
        return;
}

