/*
 * Cluster Information Service - a monitoring system for Linux clusters
 * Copyright (C) 2000 Institute of Informatics, Slovak Academy of Sciences.
 * Written by Jan Astalos (astalos.ui@savba.sk)
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * cis_info: gets basic information about cluster
 */

#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <locale.h>

#include "cis.h"
#include "cis_clnt.h"

#define DEF_INTERVAL                1.

int (*sort_function)(const void *l,const void *r);

char myhostname[MAXHOSTNAMELEN];
float interval = DEF_INTERVAL;
char buff[256];
struct cis_hostinfo *total;
int totals = 0;
int allnodes = 0;
int nodelist = 0;

struct cisinfo {
	char *host;
        CLIENT *handle;

	struct cis_info *info;

        int nhosts;
	struct cis_hostinfo *hlist;

} *cis = NULL;

enum {
        NAME,
        PERF_MAX,
	PERF_AVAIL,
	MEM_MAX,
	MEM_AVAIL,
	SWAP_MAX,
	SWAP_AVAIL,
        SWAPRATE,
	DISKRATE,
	PROC,
	CTX,
	LOAD1,
	LOAD5,
	LOAD15,
	RELIAB,
};

struct column_info {
	int num;
	int indent;
	int print;
	int width;
} columns[] = {
	{NAME,      -1, 1},
	{PERF_MAX,   1, 1},
	{PERF_AVAIL, 1, 1},
	{MEM_MAX,    1, 1},
	{MEM_AVAIL,  1, 1},
	{SWAP_MAX,   1, 1},
	{SWAP_AVAIL, 1, 1},
	{SWAPRATE,   0, 0},
	{DISKRATE,   0, 0},
	{PROC,       1, 1},
	{CTX,        1, 0},
	{LOAD1,      1, 1},
	{LOAD5,      1, 1},
	{LOAD15,     1, 1},
	{RELIAB,     0, 0},
};

struct header_info {
	char *text;
	int pos;
	int ncols;
} header1[] = {
	{"Performance",  PERF_MAX, 2},
	{"RAM [KB]",     MEM_MAX,  2},
	{"Swap [KB]",    SWAP_MAX, 2},
	{"Swap",         SWAPRATE, 1},
	{"Disk",         DISKRATE, 1},
	{"Average load", LOAD1,    3},
}, header2[] = {
	{"Name",        NAME,        1},
	{" max ",       PERF_MAX,    1},
	{"avail",       PERF_AVAIL,  1},
	{"max",         MEM_MAX,     1},
	{"avail",       MEM_AVAIL,   1},
	{"max",         SWAP_MAX,    1},
	{"avail",       SWAP_AVAIL,  1},
	{"in/out",      SWAPRATE,    1},
	{" r/w ",       DISKRATE,    1},
	{"Proc",        PROC,        1},
	{"Ctx",         CTX,         1},
	{"1min",        LOAD1,       1},
	{"5min",        LOAD5,       1},
	{"15min",       LOAD15,      1},
	{"Reliability", RELIAB,      1},
};
	

#define for_each_column(col) for (col = columns; (char*) col < (char*)&columns+sizeof(columns); col++)

#define for_each_host(hi) for (hi = cis->hlist; hi < cis->hlist+cis->nhosts; hi++)

#define width(type, fmt, args...) \
	len = sprintf(buff, fmt, args); \
	if (len > columns[type].width) \
	        columns[type].width = len;

#define getwidth(type, fmt, args...) \
	if (columns[type].print) \
	        for_each_host(hi) { \
		        if (!allnodes && hi->CPU_available == 0) \
			        continue; \
	                len = sprintf(buff, fmt, args); \
	                if (len > columns[type].width) \
	                         columns[type].width = len; \
                }


#define for_each_header(h, tab) for (h = tab; (char*) h < (char*)&tab+sizeof(tab); h++)

void repeat(char c, int length)
{
	for (; length; length--)
		printf ("%c", c);
}

#define blank(len) repeat(' ', len)

void hprint(struct header_info *h)
{
	int i, w, len;

	if (!columns[h->pos].print)
		return;
	
	for (i = h->pos, w = -1; i < h->pos+h->ncols; i++)
		w += columns[i].width+1;

	len = w - strlen(h->text);

	if (columns[h->pos].indent != -1)
		blank(len - len/2);

	printf("%s", h->text);
	if (columns[h->pos].indent != -1)
		blank(len/2 + 1);
	else
		blank(len+1);
}

void cprint(int type, char *fmt, ...)
{
	va_list ap;
	int len;

	if (!columns[type].print)
		return;

	va_start(ap, fmt);

	len = vsprintf(buff, fmt, ap);
	len = columns[type].width - len;

	switch (columns[type].indent) {
	case -1:
		printf("%s", buff);
		blank(len+1);
		break;
	case 1:
		blank(len);
		printf("%s", buff);
		blank(1);
		break;
	case 0:
		blank(len - len / 2);
		printf("%s", buff);
		blank(len/2 + 1);
	}
}

int sort_name(const void *left, const void *right)
{
	struct cis_hostinfo *l = (struct cis_hostinfo *) left;
        struct cis_hostinfo *r = (struct cis_hostinfo *) right;

	return (strcmp(l->name, r->name));
}

int sort_perf(const void *left, const void *right)
{
	struct cis_hostinfo *l = (struct cis_hostinfo *) left;
        struct cis_hostinfo *r = (struct cis_hostinfo *) right;

	return ((int) (r->CPU_available/1e2*r->performance -
		       l->CPU_available/1e2*l->performance));
}

int sort_mem(const void *left, const void *right)
{
	struct cis_hostinfo *l = (struct cis_hostinfo *) left;
        struct cis_hostinfo *r = (struct cis_hostinfo *) right;

	return ((int) (r->freeram+r->bufferram+r->cachedram -
		       (l->freeram+l->bufferram+l->cachedram)));
}

#define sum(type) hi->type[0]+hi->type[1]+hi->type[2]+hi->type[3]

int main (int argc,char *argv[])
{
	struct cis_hostinfo *hi;
	struct timeval current_time;
        char *nodes = NULL;
	int c, pos, len, nnodes = 0;
	struct column_info *col;
	struct header_info *h;
	float perf_av = 0.;

	setlocale(LC_ALL, "");
	setlocale(LC_NUMERIC, "C");

	if (gethostname((char *)& myhostname, MAXHOSTNAMELEN) < 0) {
		fprintf(stderr, "cannot get my hostname\n");
		exit(1);
	}

	cis = calloc(1, sizeof (struct cisinfo));
	if (!cis) {
		fprintf(stderr, "not enough memory\n");
		exit(1);
	}
        cis->host = getenv("CIS_SERVER");

	total = calloc(1, sizeof (struct cis_hostinfo));
	if (!total) {
		fprintf(stderr, "not enough memory\n");
		exit(1);
	}

        nodes = getenv("CIS_NODES");

	sort_function = sort_name;
	
	while ((c = getopt(argc, argv, "i:s:cdrwtalPM")) != EOF) {

		switch (c) {
		case 'i':
			interval = atof(optarg);
			if (interval == 0.) {
				fprintf(stderr, "interval cannot be set to zero. reset to default %5.2f [s]\n", DEF_INTERVAL);
				interval = DEF_INTERVAL;
			}
			break;
		case 's':
			cis->host = strdup(optarg);  break;
		case 'r':
			columns[RELIAB  ].print = 1; break;
		case 'w':
			columns[SWAPRATE].print = 1; break;
		case 'c':
			columns[CTX     ].print = 1; break;
		case 'd':
			columns[DISKRATE].print = 1; break;
		case 't':
			totals = 1; break;
		case 'a':
			allnodes = 1; break;
		case 'l':
			nodelist = 1; break;
		case 'P':
			sort_function = sort_perf; break;
		case 'M':
			sort_function = sort_mem; break;
		case ':':
                case '?':
			fprintf(stderr, "usage: cis_info [-s cis_server] [-cdrwtaPM]\n");
			exit(1);
		}
        }

	if (!cis->host)
		cis->host = strdup(myhostname);

	cis->handle = cisConnect(cis->host);
	if (!cis->handle) {
		fprintf(stderr, "cannot connect to cis server %s\n", cis->host);
		exit(1);
	}

	cis->info = cisInfo(cis->handle);
        if (!cis->info) {
		fprintf(stderr, "cannot get information from cis server %s\n", cis->host);
                exit(1);
        }

        if (!nodes || allnodes) {
                cis->nhosts = cisHostlist(cis->handle, &cis->hlist, 0, "", interval);
                if (cis->nhosts < 1) {
                        fprintf(stderr, "no hosts\n");
                        exit(1);
                }
        } else {
                char *s = strtok(nodes, ":");
                do {
                        hi = cisHostinfo(cis->handle, s, "", interval);
                        cis->hlist = realloc(cis->hlist, (cis->nhosts+1) * sizeof(struct cis_hostinfo));
                        if (!cis->hlist) {
                                fprintf(stderr, "not enough memory\n");
                                exit(1);
                        }
                        cis->hlist[cis->nhosts++] = *hi;
                        free(hi);

                } while (s = strtok(NULL, ":"));
        }
        if (nodelist) {
                for_each_host(hi) {
                        if (!allnodes && hi->CPU_available == 0)
                                continue;
                        printf("%s ", hi->name);
                }
                exit(0);
        }

	gettimeofday(&current_time, NULL);

	for_each_column(col)
		col->width = strlen(header2[col-columns].text);

	if (totals) {
		for_each_host(hi) {
			if (!allnodes && hi->CPU_available == 0)
				continue;
			nnodes++;
                        perf_av += hi->performance ? hi->CPU_available/1e4*hi->performance : 0;
                        total->performance += hi->performance;
			total->totalram    += hi->totalram;
			total->freeram     += hi->freeram+hi->bufferram+hi->cachedram;
			total->totalswap   += hi->totalswap;
			total->freeswap    += hi->freeswap;
			total->procnum     += hi->procnum;
			total->loads[0]    += hi->loads[0];
			total->loads[1]    += hi->loads[1];
			total->loads[2]    += hi->loads[2];
		}
		
                total->CPU_available = total->performance ? perf_av/total->performance*1e4 : 0;
		total->loads[0] /= nnodes;
		total->loads[1] /= nnodes;
		total->loads[2] /= nnodes;
	}
	
	getwidth (NAME      , "%s",   hi->name);
	getwidth (PERF_MAX  , "%.2f", hi->performance);
        getwidth (PERF_AVAIL, "%.2f", hi->performance ? hi->CPU_available/1e4*hi->performance : 0);
	getwidth (MEM_MAX   , "%ld",  hi->totalram);
	getwidth (MEM_AVAIL , "%ld",  hi->freeram+hi->bufferram+hi->cachedram);
	getwidth (SWAP_MAX  , "%ld",  hi->totalswap);
	getwidth (SWAP_AVAIL, "%ld",  hi->freeswap);
	getwidth (SWAPRATE  , "%ld/%ld", hi->swpin, hi->swpout);
	getwidth (DISKRATE  , "%ld/%ld", hi->disks_read, hi->disks_write);
	getwidth (PROC      , "%d",   hi->procnum);
	getwidth (CTX       , "%ld",  hi->ctx_swtch);
	getwidth (LOAD1     , "%.2f", hi->loads[0]/100.);
	getwidth (LOAD5     , "%.2f", hi->loads[1]/100.);
	getwidth (LOAD15    , "%.2f", hi->loads[2]/100.);
	getwidth (RELIAB    , "%s",   rel_types[hi->reliability]);

	if (totals) {
		width (PERF_MAX  , "%.2f", total->performance);
                width (PERF_AVAIL, "%.2f", total->performance ? total->CPU_available/1e4*total->performance : 0);
		width (PERF_AVAIL, "%.2f", total->CPU_available/1e4*total->performance);
		width (MEM_MAX   , "%ld",  total->totalram);
		width (MEM_AVAIL , "%ld",  total->freeram);
		width (SWAP_MAX  , "%ld",  total->totalswap);
		width (SWAP_AVAIL, "%ld",  total->freeswap);
		width (PROC      , "%d",   total->procnum);
		width (LOAD1     , "%.2f", total->loads[0]/100.);
		width (LOAD5     , "%.2f", total->loads[1]/100.);
		width (LOAD15    , "%.2f", total->loads[2]/100.);
	}
	printf("\n");
	printf("Cluster: %s\n", cis->info->name ? cis->info->name : "<noname>");
	printf("\n");

	pos = 0;
	for_each_header(h, header1) {
		for (col = columns+pos; col < columns+h->pos; col++)
			if (col->print)
				blank(col->width+1);
		hprint(h);
		pos = h->pos+h->ncols;
	}
	printf("\n");

	pos = 0;
	for_each_header(h, header2) {
		hprint(h);
		pos = h->pos+h->ncols;
	}
	printf("\n");

	for_each_column(col)
		if (col->print)
			repeat('-', col->width+1);
	printf("\n");

	qsort(cis->hlist, cis->nhosts, sizeof(struct cis_hostinfo), sort_function);
	
	for_each_host(hi) {
                if (!allnodes && hi->CPU_available == 0)
			continue;
                cprint (NAME      , "%s",      hi->name);
/*                if (hi->CPU_available == 0) {
                        printf ("not available\n");
			continue;
		}*/
		cprint (PERF_MAX  , "%.2f",    hi->performance);
                cprint (PERF_AVAIL, "%.2f",    hi->performance ? hi->CPU_available/1e4*hi->performance : 0);
		cprint (MEM_MAX   , "%ld",     hi->totalram);
		cprint (MEM_AVAIL , "%ld",     hi->freeram+hi->bufferram+hi->cachedram);
		cprint (SWAP_MAX  , "%ld",     hi->totalswap);
		cprint (SWAP_AVAIL, "%ld",     hi->freeswap);
		cprint (SWAPRATE  , "%ld/%ld", hi->swpin, hi->swpout);
		cprint (DISKRATE  , "%ld/%ld", hi->disks_read, hi->disks_write);
		cprint (PROC      , "%d",      hi->procnum);
		cprint (CTX       , "%ld",     hi->ctx_swtch);
		cprint (LOAD1     , "%.2f",    hi->loads[0]/100.);
		cprint (LOAD5     , "%.2f",    hi->loads[1]/100.);
		cprint (LOAD15    , "%.2f",    hi->loads[2]/100.);
		cprint (RELIAB    , "%s",      rel_types[hi->reliability]);
		printf("\n");

	}
	for_each_column(col)
		if (col->print)
			repeat('-', col->width+1);
	printf("\n");

	if (!totals)
		exit(0);
	
	cprint (NAME      , "%s"  ,"Total");
	cprint (PERF_MAX  , "%.2f", total->performance);
        cprint (PERF_AVAIL, "%.2f", total->performance ? total->CPU_available/1e4*total->performance : 0);
	cprint (MEM_MAX   , "%ld",  total->totalram);
	cprint (MEM_AVAIL , "%ld",  total->freeram);
	cprint (SWAP_MAX  , "%ld",  total->totalswap);
	cprint (SWAP_AVAIL, "%ld",  total->freeswap);
	cprint (SWAPRATE  , "");
	cprint (DISKRATE  , "");
	cprint (PROC      , "%ld",  total->procnum);
	cprint (CTX       , "");
	cprint (LOAD1     , "%.2f", total->loads[0]/100.);
	cprint (LOAD5     , "%.2f", total->loads[1]/100.);
	cprint (LOAD15    , "%.2f", total->loads[2]/100.);
	printf("\n");

	exit(0);
}

